import colorsys
import random
from enum import Enum
from typing import Dict, List, Tuple, Union

import matplotlib.colors as mplc
import matplotlib.pyplot as plt
import numpy as np
from matplotlib.patches import Polygon, Rectangle
from PIL import Image
from scipy import special
from skimage import measure


def change_color_brightness(color: Tuple, brightness_factor: float):
    """
    Depending on the brightness_factor, gives a lighter or darker color i.e. a color with
    less or more saturation than the original color.
    Args:
        color: color of the polygon. Refer to `matplotlib.colors` for a full list of
            formats that are accepted.
        brightness_factor (float): a value in [-1.0, 1.0] range. A lightness factor of
            0 will correspond to no change, a factor in [-1.0, 0) range will result in
            a darker color and a factor in (0, 1.0] range will result in a lighter color.
    Returns:
        modified_color (tuple[double]): a tuple containing the RGB values of the
            modified color. Each value in the tuple is in the [0.0, 1.0] range.
    """
    assert brightness_factor >= -1.0 and brightness_factor <= 1.0
    color = mplc.to_rgb(color)
    polygon_color = colorsys.rgb_to_hls(*mplc.to_rgb(color))
    modified_lightness = polygon_color[1] + (brightness_factor *
                                             polygon_color[1])
    modified_lightness = 0.0 if modified_lightness < 0.0 else modified_lightness
    modified_lightness = 1.0 if modified_lightness > 1.0 else modified_lightness
    modified_color = colorsys.hls_to_rgb(polygon_color[0], modified_lightness,
                                         polygon_color[2])
    return modified_color


def draw_text(ax: plt.Axes,
              text: str,
              position: Tuple,
              font_size: float,
              color: str = "g",
              horizontal_alignment: str = "center",
              rotation: int = 0):
    """
    Args:
        text (str): class label
        position (tuple): a tuple of the x and y coordinates to place text on image.
        font_size (int, optional): font of the text. If not provided, a font size
            proportional to the image width is calculated and used.
        color: color of the text. Refer to `matplotlib.colors` for full list
            of formats that are accepted.
        horizontal_alignment (str): see `matplotlib.text.Text`
        rotation: rotation angle in degrees CCW
    Returns:
        output (VisImage): image object with text drawn.
    """
    # since the text background is dark, we don't want the text to be dark
    color = np.maximum(list(mplc.to_rgb(color)), 0.2)
    color[np.argmax(color)] = max(0.8, np.max(color))

    x, y = position
    ax.text(
        x,
        y,
        text,
        size=font_size * 1,
        family="sans-serif",
        bbox={
            "facecolor": "black",
            "alpha": 0.8,
            "pad": 0.7,
            "edgecolor": "none"
        },
        verticalalignment="top",
        horizontalalignment=horizontal_alignment,
        color=color,
        zorder=10,
        rotation=rotation,
    )


class BoxType(Enum):
    xyxy = 1
    xywh = 2


def draw_instances(img: Union[Image.Image, np.ndarray],
                   boxes: Union[np.ndarray, List],
                   labels: Union[np.ndarray, List],
                   scores: Union[np.ndarray, List],
                   masks: Union[np.ndarray, List],
                   idx_class_dict: Dict[int, str],
                   title: str = '',
                   figsize: Tuple = (16, 8),
                   show_boxes: bool = False,
                   show_masks: bool = True,
                   min_score: float = 0.5,
                   min_area: int = 0,
                   colors: List = None,
                   ax: plt.Axes = None,
                   box_type: BoxType = BoxType.xyxy,
                   only_class_idxs: List[int] = None):
    """draw the instances from a object detector or an instance segmentation model

    Args:
        img (np.ndarray): an image with shape (width, height, channels)
        boxes (np.ndarray): an array of shape (nboxes, 4)
        labels (np.ndarray): an array of shape (nlabels,)
        scores (np.ndarray): an array of shape (nscores,)
        masks (np.ndarray): an array of shape [nmasks, 1, width, height ]
        idx_class_dict (Dict[int, str]): a dictionary that maps class id to class name
        title (str, optional): [description]. Defaults to ''.
        figsize (Tuple, optional): [description]. Defaults to (16, 8).
        show_boxes (bool, optional): [description]. Defaults to False.
        show_masks (bool, optional): [description]. Defaults to True.
        min_score (float, optional): [description]. Defaults to 0.5.
        colors (List, optional): [description]. Defaults to None.
        ax (plt.Axes, optional): [description]. Defaults to None.
        box_type (BoxType, optional): [description]. Defaults to BoxType.xyxy.
        only_class_idxs (List[int], optional): [description]. Defaults to None.

    Returns:
        [type]: [description]
    """

    if boxes is not None and len(boxes) > 0 and len(np.array(boxes).shape) != 2:
        raise ValueError(
            f'the shape of the boxes should be (N_BOXES, 4) while shape is {np.array(boxes).shape}'
        )

    if masks is not None and len(masks) > 0 and len(np.array(masks).shape) < 3:
        raise ValueError(
            f'the shape of the masks should be (N_MASKS, HEIGHT, WIDTH) while shape is {np.array(masks).shape}'
        )

    labels_names = create_text_labels(labels, scores, idx_class_dict)

    colors = generate_colormap(len(idx_class_dict) +
                               1) if colors is None else colors

    if ax is None:
        _, ax = plt.subplots(figsize=figsize)

    only_class_idxs = list(
        idx_class_dict.keys()) if only_class_idxs is None else only_class_idxs

    if isinstance(img, Image.Image):
        width, height = img.width, img.height
    else:
        height, width = img.shape[:2]

    ax.set_ylim(height + 10, -10)
    ax.set_xlim(-10, width + 10)
    ax.axis('off')
    ax.set_title(title)

    out_image = np.array(img).astype(np.uint8)

    for idx in range(len(labels)):
        label_id = labels[idx]

        if label_id not in only_class_idxs:
            continue

        label_name = labels_names[idx]
        score = scores[idx]
        if score < min_score:
            continue

        if show_masks:
            mask = np.squeeze(masks[idx, ...])
        color = colors[label_id]

        if show_boxes:
            box = boxes[idx]
            if box_type.value == BoxType.xyxy.value:
                x0, y0, x1, y1 = box
                x, y, w, h = x0, y0, x1 - x0, y1 - y0
            else:
                x, y, w, h = box

            area = w * h
            if area < min_area:
                continue

            p = Rectangle((x, y),
                          w,
                          h,
                          linewidth=2,
                          alpha=0.7,
                          linestyle="dashed",
                          edgecolor=color,
                          facecolor='none')
            ax.add_patch(p)

        # add the caption
        # draw text in the center (defined by median) when box is not drawn
        # median is less sensitive to outliers.
        if show_masks:
            text_pos = np.median(mask.nonzero(), axis=1)[::-1]
        else:
            text_pos = (x + 5, y + 5)

        horiz_align = "left"

        lighter_color = change_color_brightness(color, brightness_factor=0.7)
        font_size = 10
        draw_text(ax, label_name, text_pos, font_size, lighter_color,
                  horiz_align)
        if show_masks:
            padded_mask = np.zeros((mask.shape[0] + 2, mask.shape[1] + 2),
                                   dtype=np.float32)
            padded_mask[1:-1, 1:-1] = mask
            contours = measure.find_contours(padded_mask, 0.5)
            for verts in contours:
                # Subtract the padding and flip (y, x) to (x, y)
                verts = np.fliplr(verts) - 1
                p = Polygon(verts,
                            facecolor=color,
                            edgecolor=color,
                            fill=True,
                            alpha=.5)
                ax.add_patch(p)
    ax.imshow(out_image)
    return ax


def generate_colormap(nelems: int, scaled: bool = False, bright: bool = True):
    # Generate colors for drawing bounding boxes.
    brightness = 1. if bright else .7
    hsv_tuples = [(x / nelems, 1., brightness) for x in range(nelems)]
    colors = [colorsys.hsv_to_rgb(*x) for x in hsv_tuples]
    if scaled:
        colors = [
            (int(x[0] * 255), int(x[1] * 255), int(x[2] * 255)) for x in colors
        ]
    random.shuffle(colors)
    return colors


def create_text_labels(classes: List[int], scores: List[float],
                       idx_class_dict: Dict[int, str]):
    """
    Args:
        classes (list[int] or None):
        scores (list[float] or None):
        idx_class_dict (Dict[int, str] or None):
    Returns:
        list[str]
    """
    labels = [idx_class_dict[i] for i in classes]
    labels = [
        "{} {:.0f}%".format(label, score * 100)
        for label, score in zip(labels, scores)
    ]
    return labels


# def draw_segmentation_map(mask: np.ndarray, colors: List[str]):
#     """draw a segmentation map based on a class mask.
#     The mask have to contains number that represent the class

#     Args:
#         mask (np.ndarray): [description]
#         colors (List[str], optional): [description]. Defaults to None.

#     Returns:
#         [type]: [description]
#     """
#     colors_rgb = np.array([matplotlib.colors.to_rgb(c) for c in colors]) * 255

#     r = np.zeros_like(mask).astype(np.uint8)
#     g = np.zeros_like(mask).astype(np.uint8)
#     b = np.zeros_like(mask).astype(np.uint8)

#     for c in sorted(np.unique(mask))[1:]:
#         idx = mask == c
#         r[idx] = colors_rgb[c, 0]
#         g[idx] = colors_rgb[c, 1]
#         b[idx] = colors_rgb[c, 2]

#     rgb = np.stack([r, g, b], axis=2)
#     return rgb


def draw_segmentation(
        img: Union[np.ndarray, Image.Image],
        logits_or_mask: np.ndarray,
        idx_name_dict: Dict[int, str],
        min_conf: float,
        colors: List = None,
        title: str = '',
        ax: plt.Axes = None,
        figsize: Tuple[int, int] = (16, 8),
):
    """draw the result from a segmentation model

    Args:
        img (Union[np.ndarray, Image.Image]): an PIL image or a numpy array
        logits_or_mask (np.ndarray): it accepts:
            - the logits coming from the model with shape (n_classes, H, W), or
            - the mask coming from true annotations with shape (H,W) and containing pixel classification
        idx_name_dict (Dict[int, str]): 
        min_conf (float): the min confidence of the mask given as output
        colors (List, optional): the colors to diplay categories. Defaults to None.
        title (str, optional): [description]. Defaults to ''.
        ax (plt.Axes, optional): [description]. Defaults to None.
        figsize (Tuple[int, int], optional): [description]. Defaults to (16, 8).

    Returns:
        [plt.Axes]: the ax of the given plot 
    """
    if colors is None:
        colors = generate_colormap(len(idx_name_dict) + 1)

    if ax is None:
        _, ax = plt.subplots(figsize=figsize)

    if isinstance(img, np.ndarray):
        img = Image.fromarray(img)

    width, height = img.size
    ax.set_ylim(height + 10, -10)
    ax.set_xlim(-10, width + 10)
    ax.axis('off')
    ax.set_title(title)

    out_image = np.array(img).astype(np.uint8)

    if len(logits_or_mask.shape) == 2:
        masks = logits_or_mask
        probs = np.ones((len(idx_name_dict), *masks.shape))
    else:
        probs = special.softmax(logits_or_mask, axis=0)
        masks = probs.argmax(0)

    for cat in np.unique(masks)[1:]:
        mask = masks == cat
        conf = np.round(np.nan_to_num(probs[cat, mask].mean()), 2)
        if conf < min_conf:
            continue
        #remove all the pixels with score lower
        filt_mask = np.copy(mask)
        filt_mask[probs[cat, ...] < min_conf] = 0

        name = f'{idx_name_dict[cat]} {int(conf * 100)}%'
        color = colors[cat]

        # draw text in the center (defined by median) when box is not drawn
        # median is less sensitive to outliers.
        text_pos = np.median(filt_mask.nonzero(), axis=1)[::-1] - 20
        # horiz_align = "left"

        lighter_color = change_color_brightness(color, brightness_factor=0.7)
        font_size = 10
        draw_text(ax, name, text_pos, font_size, horizontal_alignment='left')

        padded_mask = np.zeros((filt_mask.shape[0] + 2, filt_mask.shape[1] + 2),
                               dtype=np.uint8)
        padded_mask[1:-1, 1:-1] = filt_mask
        contours = measure.find_contours(padded_mask, 0.5)
        for verts in contours:
            verts = np.fliplr(verts) - 1
            p = Polygon(
                verts,
                facecolor=color,
                edgecolor=lighter_color,    # 'black',
                fill=True,
                alpha=.5)
            ax.add_patch(p)
    ax.imshow(out_image)
    return ax



from polimorfo.utils import datautils
from pathlib import Path
import os
import pytest
import shutil

BASE_PATH = Path(__file__).parent.parent / 'data'


@pytest.fixture
def cleandir():
    to_cancel = BASE_PATH / 'datasets'
    if to_cancel.exists():
        shutil.rmtree(to_cancel)


def test_download_mnist():
    dst_path = BASE_PATH / 'mnist.zip'
    file_size = datautils.download_url(
        url='https://github.com/mlampros/DataSets/raw/master/mnist.zip',
        dst_path=dst_path)
    assert file_size == 15291135
    assert dst_path.exists()
    os.remove(dst_path)


def test_unzip_file():
    dst_path = BASE_PATH / 'mnist.zip'
    _ = datautils.download_url(
        url='https://github.com/mlampros/DataSets/raw/master/mnist.zip',
        dst_path=dst_path)
    out_paths = datautils.extract_archive(str(dst_path))
    assert out_paths[0].name == 'mnist.csv'
    for out in out_paths:
        shutil.rmtree(out.parent)
    os.remove(dst_path)


def test_download_gdrivezip():
    dst_path = BASE_PATH / 'test.tar.gz'
    out_path = datautils.download_from_gdrive(
        '1Sr1fm8PaaKQuvVpl34t6xyWty61o2psH', dst_path)
    assert out_path.name == 'test.tar.gz'
    os.remove(out_path)


@pytest.mark.usefixtures("cleandir")
def test_download_archive_gdrive():
    files = datautils.download_file(name='test.tar.gz',
                                    url='1Sr1fm8PaaKQuvVpl34t6xyWty61o2psH',
                                    cache_dir=BASE_PATH)

    print(files)
    assert len(files) > 0
    for f in files:
        if f.parent.exists():
            shutil.rmtree(f.parent)


@pytest.mark.usefixtures("cleandir")
def test_download_file_gdrive():
    files = datautils.download_file(name='requirements.txt',
                                    url='1_K6gqtZrvGWsINfkY0J6Evbd3bbWLg_0',
                                    cache_dir=BASE_PATH,
                                    extract=False)

    print(files)
    assert len(files) > 0
    for f in files:
        os.remove(f)


@pytest.mark.usefixtures("cleandir")
def test_download_file_extract():
    with pytest.raises(AssertionError):
        datautils.download_file(name='requirements.txt',
                                url='1_K6gqtZrvGWsINfkY0J6Evbd3bbWLg_0',
                                cache_dir=BASE_PATH)


@pytest.mark.usefixtures("cleandir")
def test_download_archive():
    files = datautils.download_file(
        name='mnist.zip',
        url='https://github.com/mlampros/DataSets/raw/master/mnist.zip',
        cache_dir=BASE_PATH,
        extract=True)

    print(files)
    assert len(files) > 0
    shutil.rmtree(files[0].parent)
    shutil.rmtree(files[0].parent.parent)


# @pytest.mark.usefixtures("cleandir")
# def test_download_file_github():
#     files = datautils.download_file(
#         name='densenet121_weights_tf_dim_ordering_tf_kernels.h5',
#         url=
#         'https://github.com/keras-team/keras-applications/releases/download/densenet/densenet121_weights_tf_dim_ordering_tf_kernels_notop.h5',
#         cache_dir=BASE_PATH,
#         extract=False)

#     print(files)
#     assert len(files) > 0
#     shutil.rmtree(files[0].parent)



from typing import List

import numpy as np
import scipy
from deprecated import deprecated
from scipy import special

from ..utils import maskutils
from .coco import CocoDataset

__all__ = ['SemanticCoco', 'SemanticCocoDataset']


class SemanticCoco(CocoDataset):
    """
    An extension of the coco dataset to handle the output of a semantic segmentation model

    """

    def add_annotations(self,
                        img_id: int,
                        logits: np.ndarray,
                        min_conf: float,
                        one_mask_per_class: bool = False,
                        start_index=1) -> List[int]:
        """
        add the annotation from the given logits

        Args:
            img_id (int): the id of the image to associate the annotations
            logits (np.ndarray): an array of shape [NClasses, Height, Width]
            min_conf (float): the minimum confidence used to filter the generated masks
            one_mask_per_class (bool, optional): if True save only the largest mask per class (default: False)
            cats_idxs (List, optional): A list that maps the. Defaults to None.
            start_index (int, optional): the index to start generating the coco polygons.
                Normally, 0 encodes the background. Defaults to 1.

        Raises:
            ValueError: if the shape of masks is different than 2
            ValueError: if the shape of probs is different than 3

        Returns:
            List[int]: [the idx of the annotations added]
        """

        if not isinstance(logits, np.ndarray):
            raise ValueError(
                f'the mask type should be a numpy array not a {type(logits)}')

        if len(logits.shape) != 3:
            raise ValueError('masks.shape should equal to 3')

        probs = special.softmax(logits, axis=0)
        masks = probs.argmax(0)

        annotation_ids = []
        for cat_idx in np.unique(masks)[start_index:]:
            mask = (masks == cat_idx).astype(np.uint8)
            conf = np.round(np.nan_to_num(probs[cat_idx, mask].mean()), 2)

            if conf < min_conf:
                continue

            #remove all the pixels with score lower
            filt_mask = np.copy(mask)
            filt_mask[probs[cat_idx, ...] < min_conf] = 0

            cat_id = int(cat_idx)
            if cat_id not in self.cats:
                raise ValueError(f'cats {cat_id} not in dataset categories')

            groups, n_groups = scipy.ndimage.label(filt_mask)

            if one_mask_per_class:
                largest_area = 0
                largest_id = 0

                for group_idx in range(1, n_groups + 1):
                    group_mask = (groups == group_idx).astype(np.uint8)
                    if group_mask.sum() > largest_area:
                        largest_area = group_mask.sum()
                        largest_id = group_idx

                n_groups = 1
                groups[groups != largest_id] = 0

            # get the groups starting from label 1
            for group_idx in range(1, n_groups + 1):
                group_mask = (groups == group_idx).astype(np.uint8)
                polygons = maskutils.mask_to_polygon(group_mask)
                if len(polygons) == 0:
                    continue

                bbox = maskutils.bbox(polygons, *masks.shape).tolist()
                # an exception is generated when the mask has less than 3 points
                area = int(maskutils.area(group_mask))
                if area == 0:
                    continue
                group_prob_mask = group_mask * probs[cat_idx]
                score = float(np.mean(group_prob_mask[group_prob_mask > 0]))
                annotation_ids.append(
                    self.add_annotation(img_id, cat_id, polygons, area, bbox, 0,
                                        score))
        return annotation_ids


@deprecated(version='0.9.34', reason='you should use SemanticCoco')
class SemanticCocoDataset(SemanticCoco):

    def __init__(self, coco_path: str, image_path: str = None) -> None:
        super().__init__(coco_path, image_path)



from pathlib import Path
from typing import Any, Dict, List, Union

import numpy as np

from polimorfo.datasets import CocoDataset, InstanceCoco, SemanticCocoDataset


def rename_annotation_keys(ann: Dict[str, Union[str, int, list, np.ndarray]]) -> Dict:
    """Renames keys for a Coco annotation in order to feed add_annotation method properly.
    Args:
        ann (Dict[str, Union[str, int, list, np.ndarray]]): Coco annotation
    Returns:
        Dict: same as input annotation with keys renamed.
    """
    key_map = { 
                'category_id': 'cat_id',
                'segmentation': 'segmentation',
                'area': 'area',
                'iscrowd': 'is_crowd',
                'score': 'score',
                'bbox': 'bbox'
              }
    assert set(list(ann.keys()) + ['score']) >= (set(key_map.keys())), f'{set(ann.keys())}, {set(key_map.keys())}'
    return { v: ann.get(k, None) for k,v in key_map.items() }

def get_img_meta_by_name(ds, image_name):
    try:
        return [ img_meta for img_meta in ds.imgs.values() if img_meta['file_name'] == image_name ][0]
    except:
        log.warning('image not present in dataset.')
        return None

def merge_datasets(datasets: List[Union[SemanticCocoDataset, CocoDataset, InstanceCoco]],
          dest_path: Union[str, Path] = None) -> CocoDataset:
    
    # compute category to cat_idx map for each dataset
    cat_to_idx = [ { v['name']: k for k,v in ds_item.cats.items() } for ds_item in datasets ] 
    category_names = [ list(ctx.keys()) for ctx in cat_to_idx ]
    flat_category_name = sorted([ item for sublist in category_names for item in sublist ])

    # collect all category names
    merged_cat_names = set(flat_category_name)

    merged_idx_to_cat = dict(enumerate(sorted(merged_cat_names), start=1))
    merged_cat_to_idx = { v:k for k,v in merged_idx_to_cat.items() }

    # map categories for each dataset to merged one
    cat_map = [ { v: merged_cat_to_idx[k]  for k,v in ctx.items() } for ctx in cat_to_idx ]

    # initializize merged dataset
    out_ds = CocoDataset(dest_path) 
    for idx, cat_name in sorted(merged_idx_to_cat.items()):
        out_ds.add_category(cat_name, 'thing')

    for ds_item, cat_map_item in zip(datasets, cat_map):
        img_names = { a['file_name']: a['id'] for a in out_ds.imgs.values() }
        # add images
        for img_meta in ds_item.imgs.values():
            # remark: if datasets are not disjoint the might bo overlap here to take care of, i.e. possible multiple definitions of images
            if img_meta['file_name'] not in img_names:
                img_idx = out_ds.add_image(**img_meta)
                img_names[img_meta['file_name']] = img_idx
            else:
                img_idx = img_names[img_meta['file_name']]

            orig_img_idx = img_meta['id']
            anns = ds_item.get_annotations(img_idx=orig_img_idx)
            # add anns
            for ann in anns:
                mapped_ann = rename_annotation_keys(ann)
                mapped_ann['img_id'] = img_idx # new image idx
                mapped_ann['cat_id'] = cat_map_item[mapped_ann['cat_id']] # new category index
                out_ds.add_annotation(**mapped_ann)
                
    out_ds.reindex()
    return out_ds



from pathlib import Path

import numpy as np

from polimorfo.datasets.semanticcoco import SemanticCocoDataset

BASE_PATH = Path(__file__).parent.parent / 'data'


def test_add_annotations():
    ds = SemanticCocoDataset('fake.json')

    img_id = ds.add_image(BASE_PATH / 'test_nodamage.jpg', 100, 100)
    ds.add_category('cat1', 'thing')
    ds.add_category('cat2', 'thing')

    mask_logits = np.random.rand(3, 256, 256)
    mask_logits[1:3] += 2

    ann_idx = ds.add_annotations(img_id, mask_logits, .2)

    assert len(ann_idx) > 0


def test_add_annotations_one_label_per_class():
    ds = SemanticCocoDataset('fake.json')

    img_id = ds.add_image(BASE_PATH / 'test_nodamage.jpg', 100, 100)
    ds.add_category('cat1', 'thing')
    ds.add_category('cat2', 'thing')

    mask_logits = np.random.rand(3, 256, 256)
    mask_logits[1, 0:50, 0:50] = 2

    ann_idx = ds.add_annotations(img_id,
                                 mask_logits,
                                 .2,
                                 one_mask_per_class=True)
    assert len(ann_idx) > 0


def test_add_annotations_zero_mask():
    ds = SemanticCocoDataset('fake.json')

    img_id = ds.add_image(BASE_PATH / 'test_nodamage.jpg', 100, 100)
    ds.add_category('cat1', 'thing')
    ds.add_category('cat2', 'thing')

    mask_logits = np.zeros((3, 256, 256))
    mask_logits[1:3] += 2

    ann_idx = ds.add_annotations(img_id, mask_logits, .2)

    assert len(ann_idx) == 0


def test_call_count_statistics():
    ds = SemanticCocoDataset('fake.json')

    img_id = ds.add_image(BASE_PATH / 'test_nodamage.jpg', 100, 100)
    ds.add_category('cat1', 'thing')
    ds.add_category('cat2', 'thing')

    mask_logits = np.random.rand(3, 256, 256)
    mask_logits[1:3] += 2

    ann_idx = ds.add_annotations(img_id, mask_logits, .4)
    # ds.reindex()

    print(ds.count_annotations_per_category())
    print(ds.count_images_per_category())

    assert len(ann_idx) > 0



from polimorfo.utils.mergeutils import merge_datasets, get_img_meta_by_name
from polimorfo.datasets import CocoDataset
from pathlib import Path
import numpy as np

BASE_PATH = Path(__file__).parent.parent / 'data'

def test_merge_datasets():

    datasets = [ CocoDataset(coco_path=BASE_PATH / dataset_name) for dataset_name 
                    in ['dataset1.json', 'dataset2.json'] ]

    merged_ds = merge_datasets(datasets, BASE_PATH / 'fake_merge.json') 

    assert len(merged_ds.anns) == 20

    for k in ['scratch', 'dent']:
        assert merged_ds.count_images_per_category()[k] == np.sum([ds_item.count_images_per_category()[k] for ds_item in datasets])
        assert merged_ds.count_annotations_per_category()[k] == np.sum([ds_item.count_annotations_per_category()[k] for ds_item in datasets])

    # check image meta/ anns consistency
    for ds_item in datasets:
        for img_idx, img_meta in ds_item.imgs.items():

            anns = ds_item.get_annotations(img_idx)

            merged_img_meta = get_img_meta_by_name(merged_ds, img_meta['file_name'])
            merged_img_idx = merged_img_meta['id']
            merged_anns = merged_ds.get_annotations(merged_img_idx)
            # number of ann per image
            assert len(merged_anns) == len(anns)
            # number of cats per image
            merged_cats = { a['category_id'] for a in merged_anns }
            cats = { a['category_id'] for a in anns }
            assert len(merged_cats) == len(cats)
            # number of annotated points per image
            merged_segm_points = { len(a['segmentation']) for a in merged_anns }
            segm_points = { len(a['segmentation']) for a in anns }
            assert len(merged_segm_points) == len(segm_points)
            # annotation areas per image
            merged_areas = { a['area'] for a in merged_anns }
            areas = { a['area'] for a in anns }
            assert merged_areas == areas
        


from pathlib import Path
from polimorfo.datasets.coco import CocoDataset
from polimorfo.utils import cocoeval
import numpy as np

BASE_PATH = Path(__file__).parent.parent / 'data'


def test_generate_predictions():
    ds_path = BASE_PATH / 'hair_drier_toaster_bear.json'
    df = cocoeval.generate_predictions(ds_path, ds_path)
    assert len(df) > 0
    assert df['IOU'].mean() == 1.


def test_generate_predictions_one_class():
    ds_path = BASE_PATH / 'hair_drier_toaster_bear.json'
    df = cocoeval.generate_predictions(ds_path, ds_path, category_idxs=[1])
    assert len(df) > 0
    assert df['IOU'].mean() > 0.


def test_mean_average_precision_and_recall():
    ds_path = BASE_PATH / 'hair_drier_toaster_bear.json'
    df = cocoeval.generate_predictions(ds_path, ds_path)
    map, mar = cocoeval.mean_average_precision_and_recall(df)
    assert map == 1.
    assert mar == 1.


def test_mean_average_precision_and_recall_per_class():
    ds_path = BASE_PATH / 'hair_drier_toaster_bear.json'
    df = cocoeval.generate_predictions(ds_path, ds_path)
    class_idx_metrics = cocoeval.mean_average_precision_and_recall_per_class(df)
    class_idxs = df['true_class_id'].unique()
    class_idxs = class_idxs[class_idxs > 0]
    print(class_idx_metrics)
    assert len(class_idx_metrics) == len(class_idxs)


def test_mean_average_precision_and_recall_per_class_with_name():
    ds_path = BASE_PATH / 'hair_drier_toaster_bear.json'

    ds = CocoDataset(ds_path)
    idx_class_name = {
        idx: cat_meta['name'] for idx, cat_meta in ds.cats.items()
    }

    df = cocoeval.generate_predictions(ds_path, ds_path)
    class_idx_metrics = cocoeval.mean_average_precision_and_recall_per_class(
        df, idx_class_dict=idx_class_name)
    class_idxs = df['true_class_id'].unique()
    class_idxs = class_idxs[class_idxs > 0]
    print(class_idx_metrics)
    assert len(class_idx_metrics) == len(class_idxs)


def test_mean_average_precision_and_recall_per_class_with_name_largest_range():
    ds_path = BASE_PATH / 'hair_drier_toaster_bear.json'

    ds = CocoDataset(ds_path)
    idx_class_name = {
        idx: cat_meta['name'] for idx, cat_meta in ds.cats.items()
    }

    df = cocoeval.generate_predictions(ds_path, ds_path)
    class_idx_metrics = cocoeval.mean_average_precision_and_recall_per_class(
        df, idx_class_dict=idx_class_name, range_iou=np.arange(0, 1, 0.5))
    class_idxs = df['true_class_id'].unique()
    class_idxs = class_idxs[class_idxs > 0]
    print(class_idx_metrics)
    assert len(class_idx_metrics) == len(class_idxs)


def test_mean_average_precision_and_recall_per_class_with_name_min_score():
    ds_path = BASE_PATH / 'hair_drier_toaster_bear.json'

    ds = CocoDataset(ds_path)
    idx_class_name = {
        idx: cat_meta['name'] for idx, cat_meta in ds.cats.items()
    }

    df = cocoeval.generate_predictions(ds_path, ds_path)
    class_idx_metrics = cocoeval.mean_average_precision_and_recall_per_class(
        df, idx_class_dict=idx_class_name, min_score=0.5)
    class_idxs = df['true_class_id'].unique()
    class_idxs = class_idxs[class_idxs > 0]
    print(class_idx_metrics)
    assert len(class_idx_metrics) == len(class_idxs)


def test_precision_recall_per_image():
    ds_path = BASE_PATH / 'hair_drier_toaster_bear.json'
    df = cocoeval.generate_predictions(ds_path, ds_path)

    map, mar = cocoeval.precision_recall_per_image(df, df['img_path'].loc[0])
    assert map > 0
    assert mar > 0